# This file is part of MOXελ.
# Copyright (C) 2023-2024 Antonios P. Sarikas

# MOXελ is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

r"""
This module provides helper functions for creating voxels.

.. note::
    Currently, interactions are modelled with the Lennard-Jones (LJ) potential.

.. attention::
    Consider tuning the ``n_jobs`` parameter to get the best performance
    for your system::

        from timeit import timeit

        setup = 'from moxel.utils import voxels_from_file'
        n_jobs = [1, 2, 8, 16]  # Modify this according to your system.

        for n in n_jobs:
            stmt = f'voxels_from_file("path/to/cif", n_jobs={n})'
            time = timeit(stmt=stmt, setup=setup, number=1)
            print(f'Time with {n} jobs: {time:.3f} s')
"""

import os
import itertools
from pathlib import Path
from multiprocessing import Pool
import warnings

import numpy as np
from tqdm import tqdm
from pymatgen.core import Structure

from . _params import lj_params

warnings.filterwarnings('ignore')

# Default values for voxels calculation.
GRID_SIZE = 25
CUTOFF = 10.
EPSILON = 50.
SIGMA = 2.5
CUBIC_BOX = None
N_JOBS = None


def mic_scale_factors(r, lattice_vectors):
    r"""
    Return scale factors to satisfy minimum image convention [MIC]_.

    Parameters
    ----------
    r : float
        The cutoff radius used in MIC convetion.
    lattice_vectors : array of shape (3, 3)
        The lattice vectors of the unit cell.
        Each row corresponds to a lattice vector.

    Returns
    -------
    scale_factors : array of shape (3,)
        ``scale_factors[i]`` scales ``lattice_vectors[i]``.

    References
    ----------
    .. [MIC] W. Smith, "The Minimum Image Convention in Non-Cubic MD Cells", 1989.
    """
    a, b, c = lattice_vectors
    volume = np.linalg.norm(np.dot(a, np.cross(b, c)))

    w_a = volume/np.linalg.norm(np.cross(b, c))
    w_b = volume/np.linalg.norm(np.cross(a, c))
    w_c = volume/np.linalg.norm(np.cross(a, b))

    return np.ceil(2 * r / np.array([w_a, w_b, w_c]))


class Grid:
    r"""
    A 3D energy grid over a crystal structure.

    Parameters
    ----------
    grid_size : int, default=25
        Number of grid points along each dimension.
    cutoff : float, default=10.0
        Cutoff radius (Å) for the LJ potential.
    epsilon : float, default=50.0
        Epsilon value (ε/K) of the probe atom.
    sigma : float, default=2.5
        Sigma value (σ/Å) of the probe atom.

    Attributes
    ----------
    structure : :class:`pymatgen.core.structure.Structure`
        Available only after :meth:`Grid.load_structure` has been called.
    structure_name : str
        Available only after :meth:`Grid.load_structure` has been called.
    cubic_box : float or None
        Available only after :meth:`Grid.calculate` has been called.
    voxels : array of shape (grid_size,)*3
       Available only after :meth:`Grid.calculate` has been called.
    """
    def __init__(
            self,
            grid_size=GRID_SIZE,
            *,
            cutoff=CUTOFF,
            epsilon=EPSILON,
            sigma=SIGMA
            ):
        self.grid_size = grid_size
        self.cutoff = cutoff
        self.epsilon = epsilon
        self.sigma = sigma

    def load_structure(self, pathname):
        r"""
        Load a crystal structure from a file in a format supported by
        :meth:`pymatgen.core.Structure.from_file`.

        Parameters
        ----------
        pathname : str
           Pathname to the file.
        """
        self.structure = Structure.from_file(pathname)
        self.structure_name = Path(pathname).stem

    def calculate(
            self,
            cubic_box=CUBIC_BOX,
            potential='lj',
            n_jobs=N_JOBS,
            ):
        r"""
        Iterate over the grid and return voxels.

        For computational efficiency and to assure (approximately) the same
        spatial resolution, the grid is overlayed over a supercell scaled
        according to MIC, see :func:`mic_scale_factors`.

        If lattice angles are significantly different than 90°, to avoid
        distortions set ``cubic_box``. In this case, the grid is overlayed over
        a cubic box of size ``cubic_box`` centered at the origin but periodicity
        is no longer guaranteed.

        Parameters
        ----------
        potential : str, default='lj'
            The potential used to calculate voxels. Currently, only the
            LJ potential is supported.
        cubic_box : float or None, default=None
            If ``None``, the simulation box is a supercell scaled according to
            MIC. Otherwise, cubic box of size ``cubic_box``.
        n_jobs : int, optional
            Number of jobs to run in parallel. If ``None``, then the number returned
            by ``os.cpu_count()`` is used.

        Returns
        -------
        voxels : array of shape (grid_size,)*3
        """
        self.cubic_box = cubic_box

        if cubic_box is not None:
            d = cubic_box / 2
            probe_coords = np.linspace(0 - d, 0 + d, self.grid_size, endpoint=False)  # Cartesian
            self._simulation_box = self.structure
        else:
            probe_coords = np.linspace(0, 1, self.grid_size, endpoint=False)  # Fractional
            scale = mic_scale_factors(self.cutoff, self.structure.lattice.matrix)
            self._simulation_box = self.structure * scale

        if potential == 'lj':
            # Cache LJ parameters for all atoms in the simulation box.
            self._lj_params = np.array(
                    [lj_params[atom.species_string] for atom in self._simulation_box]
                    )

            # Cache fractional coordinates since this is a slow function in pymatgen.
            self._frac_coords = self._simulation_box.frac_coords

            # Embarrassingly parallel.
            with Pool(processes=n_jobs) as p:
                energies = p.map(
                        self.lj_potential, itertools.product(*(probe_coords,)*3)
                        )

        self.voxels = np.array(energies, dtype=np.float32).reshape((self.grid_size,)*3)

        return self.voxels

    def lj_potential(self, coords):
        r"""
        Calculate LJ potential at cartesian or fractional
        coordinates.

        Parameters
        ----------
        coordinates : array_like of shape (3,)
            If ``cubic_box=None`` fractional, else cartesian.

        Returns
        -------
        energy : float
        """
        if self.cubic_box is not None:
            cartesian_coords = coords
        else:
            cartesian_coords = self._simulation_box._lattice.get_cartesian_coords(coords)

        _, r_ij, indices, _ = self._simulation_box._lattice.get_points_in_sphere(
                self._frac_coords, cartesian_coords,
                self.cutoff, zip_results=False,
                )

        # Need to check for length of r_ij because of
        # https://github.com/materialsproject/pymatgen/issues/3794

        if len(r_ij) == 0:  # No neighbor, zero energy.
            return 0.

        if np.any(r_ij < 1e-3):  # Close contact, infinite energy.
            return np.inf

        es_j = self._lj_params[indices]
        x = (0.5 * (es_j[:, 1] + self.sigma)) / r_ij
        e = 4 * np.sqrt(es_j[:, 0] * self.epsilon)

        return np.sum(e * (x**12 - x**6))


def voxels_from_file(
        cif_pathname,
        grid_size=GRID_SIZE,
        *,
        cutoff=CUTOFF,
        epsilon=EPSILON,
        sigma=SIGMA,
        cubic_box=CUBIC_BOX,
        n_jobs=N_JOBS,
        only_voxels=True,
        ):
    r"""
    Calculate and return voxels from ``.cif`` file.

    Parameters
    ----------
    cif_pathname : str
       Pathname to the ``.cif`` file.
    only_voxels : bool, default=True
        Determines ``out`` type.
        
    Returns
    -------
    out : array or :class:`Grid`
        If ``only_voxels=True`` array, else :class:`Grid`.

    See Also
    --------
    :func:`voxels_from_dir`
        For a description of the parameters.
    """
    grid = Grid(grid_size, cutoff=cutoff, epsilon=epsilon, sigma=sigma)

    grid.load_structure(cif_pathname)
    grid.calculate(cubic_box=cubic_box, n_jobs=n_jobs)

    if only_voxels:
        return grid.voxels

    return grid


def voxels_from_files(
        cif_pathnames,
        out_pathname,
        grid_size=GRID_SIZE,
        *,
        cutoff=CUTOFF,
        epsilon=EPSILON,
        sigma=SIGMA,
        cubic_box=CUBIC_BOX,
        n_jobs=N_JOBS,
        ):
    r"""
    Calculate voxels from a list of ``.cif`` files and store them.

    Voxels are stored under ``out_pathname`` as ``.npy`` files.

    Parameters
    ----------
    cif_pathnames : list
       List of pathnames to the ``.cif`` files.
    out_pathname : str
        Pathname to the directory under which voxels are stored.

    See Also
    --------
    :func:`voxels_from_dir`
        For a description of the parameters.

    Notes
    -----
    Structures that can't be processsed are omitted.
    """
    os.mkdir(out_pathname)

    for file in tqdm(cif_pathnames, desc='Creating energy voxels'):
        try:
            name = Path(file).stem  # Name of the structure.
            grid = voxels_from_file(
                    cif_pathname=file,
                    grid_size=grid_size,
                    cutoff=cutoff,
                    epsilon=epsilon,
                    sigma=sigma,
                    cubic_box=cubic_box,
                    n_jobs=n_jobs,
                    )

            pathname = os.path.join(out_pathname, name)
            np.save(pathname, grid)  # .npy extension is appended by default.
        except Exception as e:
            print(e)


def voxels_from_dir(
        cif_dirname: str,
        out_pathname: str,
        grid_size: int = GRID_SIZE,
        *,
        cutoff: float = CUTOFF,
        epsilon: float = EPSILON,
        sigma: float = SIGMA,
        cubic_box: float | None = CUBIC_BOX,
        n_jobs: int | None = N_JOBS,
        ):
    r"""
    Calculate voxels from a directory of ``.cif`` files and store them.

    Voxels are stored under ``out_pathname`` as ``.npy`` files.

    Parameters
    ----------
    cif_dirname : str
       Pathname to the directory containing the ``.cif`` files.
    out_pathname : str
        Pathname to the directory under which voxels are stored.
    grid_size : int, default=25
        Number of grid points along each dimension.
    cutoff : float, default=10.0
        Cutoff radius (Å) for the LJ potential.
    epsilon : float, default=50.0
        Epsilon value (ε/K) of the probe atom.
    sigma : float, default=2.5
        Sigma value (σ/Å) of the probe atom.
    cubic_box : float or None, default=None
        If ``None``, the simulation box is a supercell scaled according to
        MIC. Otherwise, cubic box of size ``cubic_box``.
    n_jobs : int, optional
        Number of jobs to run in parallel. If ``None``, then the number returned
        by ``os.cpu_count()`` is used.

    Notes
    -----
    Structures that can't be processsed are omitted.
    """
    cif_pathanmes = [os.path.join(cif_dirname, f) for f in os.listdir(cif_dirname)]

    voxels_from_files(
            cif_pathanmes, out_pathname,
            grid_size=grid_size,
            cutoff=cutoff,
            epsilon=epsilon,
            sigma=sigma,
            cubic_box=cubic_box,
            n_jobs=n_jobs,
            )



# This file is part of MOXελ.
# Copyright (C) 2023 Antonios P. Sarikas

# MOXελ is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
import tempfile


class TestCLI(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory(dir='/tmp')
        self.cif_dirname = 'tests/CIFs'
        self.out_pathname = os.path.join(self.tempdir.name, 'voxels_data')

    def run_command(self, command):
        exit_status = os.system(command)
        if exit_status != 0:
            raise RuntimeError(command)

    def test_cli_run(self):
        self.run_command(
                'moxel --config tests/config.yaml'
                f' {self.cif_dirname}'
                f' {self.out_pathname}'
                ' --grid_size 3'
                )

    def tearDown(self):
        self.tempdir.cleanup()


if __name__ == '__main__':
    unittest.main()



# This file is part of MOXελ.
# Copyright (C) 2023 Antonios P. Sarikas

# MOXελ is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
import tempfile
from pathlib import Path

import numpy as np
from pymatgen.core import Structure
from moxel.utils import (
        Grid, mic_scale_factors, voxels_from_file,
        voxels_from_files, voxels_from_dir
        )


class TestUtils(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory(dir='/tmp')

    def test_grid(self):
        cif_pathname = 'tests/CIFs/IRMOF-1.cif'
        grid_size = 5
        epsilon = 40
        sigma = 1e4  # For repulsive energies to go inf.
        cutoff = 12
        cubic_box = 35

        grid = Grid(grid_size=grid_size, epsilon=epsilon, cutoff=cutoff, sigma=sigma)
        grid.load_structure(cif_pathname)
        grid.calculate(cubic_box=cubic_box)

        # Check that the name of the structure is correct.
        self.assertEqual(grid.structure_name, 'IRMOF-1')

        # Check that the voxels have the correct shape.
        self.assertEqual(grid.voxels.shape, (grid_size,)*3)

        # Check that attributes are correctly set.
        self.assertEqual(grid.grid_size, grid_size)
        self.assertEqual(grid.epsilon, epsilon)
        self.assertEqual(grid.sigma, sigma)
        self.assertEqual(grid.cutoff, cutoff)
        self.assertEqual(grid.cubic_box, cubic_box)

        # Check that values are filled properly.
        self.assertEqual(grid.voxels.max(), np.inf)

    def test_mic_scale_factors(self):
        # Load a structure that must be scaled.
        cif_pathname = 'tests/CIFs/ZnMOF-74.cif'
        structure = Structure.from_file(cif_pathname)
        scale_factors = mic_scale_factors(10, structure.lattice.matrix)

        # At least one scale factor should be different than 1.
        self.assertTrue(np.any(scale_factors != 1))

        # Load a structure that shouldn't be scaled.
        cif_pathname = 'tests/CIFs/IRMOF-1.cif'
        structure = Structure.from_file(cif_pathname)
        scale_factors = mic_scale_factors(10, structure.lattice.matrix)

        # All scale factors should be 1.
        self.assertTrue(np.all(scale_factors == 1))

    def test_voxels_from_file(self):
        cif_pathname = 'tests/CIFs/IRMOF-1.cif'
        grid_size = 5
        cutoff = 9
        epsilon = 49
        sigma = 2
        cubic_box = 25

        out = voxels_from_file(
                cif_pathname, grid_size=grid_size,
                cutoff=cutoff, epsilon=epsilon,
                sigma=sigma, cubic_box=cubic_box,
                only_voxels=False
                )

        # Check that arguments are properly passed.
        self.assertTrue(isinstance(out, Grid))
        self.assertEqual(out.grid_size, grid_size)
        self.assertEqual(out.cutoff, cutoff)
        self.assertEqual(out.epsilon, epsilon)
        self.assertEqual(out.sigma, sigma)
        self.assertEqual(out.cubic_box, cubic_box)

        # Check that output shape is correct.
        self.assertEqual(out.voxels.shape, (grid_size,)*3)

    def test_voxels_from_files(self):
        # The test assumes all files are processable.
        grid_size = 5
        cif_dir = 'tests/CIFs'
        cif_files = [f'{cif_dir}/{i}' for i in os.listdir(cif_dir)]
        names = [Path(i).stem for i in cif_files]

        # Calculate voxels and store them.
        out_pathname = f'{self.tempdir.name}/voxels_data'
        voxels_from_files(cif_files, out_pathname, grid_size=grid_size)

        npy_files = [os.path.join(out_pathname, f) for f in os.listdir(out_pathname)]

        # Check that all files are processed properly.
        for f in npy_files:
            voxels  = np.load(f)
            self.assertTrue(Path(f).stem in names)
            self.assertTrue(voxels.shape, (grid_size,)*3)

    def test_voxels_from_dir(self):
        # The test assumes all files are processable.
        cif_dir = 'tests/CIFs'
        grid_size = 5

        # Calculate voxels and store them.
        out_pathname = f'{self.tempdir.name}/voxels_data'
        voxels_from_dir(cif_dir, out_pathname, grid_size=grid_size)

        # Check that all files are processed.
        self.assertEqual(len(os.listdir(out_pathname)), len(os.listdir(cif_dir)))

    def test_load_file(self):
        cif_dir = 'tests/CIFs'

        grid = Grid()
        grid.load_structure(f'{cif_dir}/IRMOF-1.cif')

        grid = Grid()
        grid.load_structure(f'{cif_dir}/MnH28C26(N2Cl)2.json')

    def tearDown(self):
        self.tempdir.cleanup()


if __name__ == '__main__':
    unittest.main()



# This file is part of MOXελ.
# Copyright (C) 2023-2024 Antonios P. Sarikas

# MOXελ is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
MOXελ is a Python package for parallel calculation of energy voxels, with
emphasis on reticular chemistry.
"""

__author__ = 'Antonios P. Sarikas'
__copyright__ = "Copyright (C) 2023-2024 Antonios P. Sarikas"
__license__ = 'GPL-3.0-only'



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import sys, os
from importlib.metadata import version as get_version

sys.path.insert(0, os.path.abspath('../../src'))

project = 'MOXελ'
copyright = '2023-2024, Antonios P. Sarikas'
author = 'Antonios P. Sarikas'
release = get_version('pymoxel')

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
        'sphinx.ext.autodoc',
        'sphinx.ext.viewcode',
        'sphinx.ext.napoleon',
        'sphinx.ext.intersphinx',
        'sphinx_copybutton',
        'sphinx_design',
        'sphinx_issues',
        ]

#templates_path = ['_templates']
#exclude_patterns = ['modules.rst']
autodoc_typehints = 'none'

# Exclude input prompts from copybutton
copybutton_exclude = '.linenos, .gp, .go'

# The package is too small.
#intersphinx_mapping = {
#        'python': ('https://docs.python.org/3', None),
#        'numpy': ('https://numpy.org/doc/stable/', None),
#        }

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_logo = 'images/moxel_logo.svg'
html_theme_options = {'logo_only': True, 'version_selector': False}
html_static_path = ['_static']
html_css_files = ['custom.css']

# Path to GitHub repo {group}/{project}  (note that `group` is the GitHub user or organization)
issues_github_path = "adosar/moxel"

# which is the equivalent to:
issues_uri = "https://github.com/{group}/{project}/issues/{issue}"
issues_prefix = "#"
issues_pr_uri = "https://github.com/{group}/{project}/pull/{pr}"
issues_pr_prefix = "#"
issues_commit_uri = "https://github.com/{group}/{project}/commit/{commit}"
issues_commit_prefix = "@"
issues_user_uri = "https://github.com/{user}"
issues_user_prefix = "@"



# This file is part of MOXελ.
# Copyright (C) 2023-2024 Antonios P. Sarikas

# MOXελ is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

r"""
This module provides helper functions for the CLI.
"""

from jsonargparse import auto_cli

from .utils import voxels_from_dir


def moxel_cli():
    r"""CLI for the moxel package."""
    auto_cli(
            voxels_from_dir,
            epilog='For more information 👉 https://github.com/adosar/moxel',
            )
